import 'dart:convert';
import 'dart:typed_data';
import 'package:flutter/material.dart';
import 'package:btlib/btlib.dart';

Btlib driver = Btlib();

class BTDevice {
  final String name;
  final String address;
  bool connected;
  Map<String, BTService> services = Map();
  BTDevice(this.name, this.address, this.connected);
}

class BTService {
  final BTDevice device;
  final String uuid;
  Map<String, BTCharacteristic> characteristics = Map();
  BTService(this.device, this.uuid);
}

class BTCharacteristic {
  final BTService serivce;
  final String uuid;
  BTCharacteristic(this.serivce, this.uuid);
}

class DeviceTile extends StatelessWidget {
  final BTDevice device;

  DeviceTile({Key key, this.device}) : super(key: key);

  _onGetRSSI() {
    driver.getRSSI(device.address);
  }

  _onConnectDisconnect() {
    if (device.connected)
      driver.disconnect(device.address);
    else
      driver.connect(device.address);
  }

  Widget _buildTitle(BuildContext context) {
    if (device.name != null) {
      return Column(
        mainAxisAlignment: MainAxisAlignment.start,
        crossAxisAlignment: CrossAxisAlignment.start,
        children: <Widget>[
          Text(
            device.name,
            style: Theme.of(context).textTheme.title,
          ),
          Text(
            device.address,
            style: Theme.of(context).textTheme.body1,
          ),
        ],
      );
    } else {
      return Text(device.address.toString());
    }
  }

  List<Widget> _buildServiceTiles() {
    List<ServiceTile> services = List();
    device.services.forEach((key, val) {
      List<CharacteristicTile> characteristics = List();
      val.characteristics.forEach((key, val) {
        characteristics.add(CharacteristicTile(characteristic: val));
      });
      services.add(ServiceTile(
        service: val,
        characteristicTiles: characteristics,
      ));
    });
    return services.toList();
  }

  @override
  Widget build(BuildContext context) {
    return ExpansionTile(
      title: _buildTitle(context),
      trailing: Row(
        mainAxisSize: MainAxisSize.min,
        children: [
          IconButton(
            icon: Icon(Icons.wifi),
            onPressed: _onGetRSSI,
          ),
          RaisedButton(
            child: Text(device.connected ? '斷開連線' : '開始連線'),
            color: device.connected ? Colors.red : Colors.blue,
            textColor: Colors.white,
            onPressed: _onConnectDisconnect,
          ),
        ],
      ),
      children: _buildServiceTiles(),
    );
  }
}

class ServiceTile extends StatelessWidget {
  final BTService service;
  final List<CharacteristicTile> characteristicTiles;

  ServiceTile({Key key, this.service, this.characteristicTiles})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    if (characteristicTiles.length > 0) {
      return ExpansionTile(
        title: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.start,
          children: <Widget>[
            const Text('Service'),
            Text(service.uuid,
                style: Theme.of(context)
                    .textTheme
                    .body1
                    .copyWith(color: Theme.of(context).textTheme.caption.color))
          ],
        ),
        children: characteristicTiles,
      );
    } else {
      return ListTile(
        title: const Text('Service'),
        subtitle: Text(service.uuid),
      );
    }
  }
}

class CharacteristicTile extends StatelessWidget {
  final BTCharacteristic characteristic;

  CharacteristicTile({
    Key key,
    this.characteristic,
  }) : super(key: key);

  Future<String> _asyncInputDialog(BuildContext context) async {
    String txt = '';
    return showDialog<String>(
      context: context,
      barrierDismissible:
          false, // dialog is dismissible with a tap on the barrier
      builder: (BuildContext context) {
        return AlertDialog(
          title: Text('Enter text to write'),
          content: Row(
            children: <Widget>[
              Expanded(
                  child: TextField(
                autofocus: true,
                decoration: InputDecoration(
                    labelText: 'Input Text', hintText: '[text]\\n or 0x[hex]'),
                onChanged: (value) {
                  txt = value;
                },
              ))
            ],
          ),
          actions: <Widget>[
            FlatButton(
              child: Text('Ok'),
              onPressed: () {
                Navigator.of(context).pop(txt);
              },
            ),
          ],
        );
      },
    );
  }

  _onReadPressed() {
    driver.readCharacteristic(characteristic.serivce.device.address,
        characteristic.serivce.uuid, characteristic.uuid);
  }

  _onWritePressed(String txt) {
    if (txt.startsWith("0x")) {
      driver.writeCharacteristic(
          characteristic.serivce.device.address,
          characteristic.serivce.uuid,
          characteristic.uuid,
          hexToBytes(txt.substring(2)));
    } else {
      driver.writeCharacteristic(
          characteristic.serivce.device.address,
          characteristic.serivce.uuid,
          characteristic.uuid,
          Utf8Codec().encode(txt + "\n"));
    }
  }

  _onNotificationON() {
    driver.setCharacteristicNotification(
      characteristic.serivce.device.address,
      characteristic.serivce.uuid, 
      characteristic.uuid, 
      true);
  }

  _onNotificationOFF() {
    driver.setCharacteristicNotification(
      characteristic.serivce.device.address,
      characteristic.serivce.uuid, 
      characteristic.uuid, 
      false);
  }

  List<int> hexToBytes(String hex) {
    const String _ALPHABET = "0123456789abcdef";
    String str = hex.replaceAll(" ", "");
    str = str.toLowerCase();
    if (str.length % 2 != 0) {
      str = "0" + str;
    }
    Uint8List result = new Uint8List(str.length ~/ 2);
    for (int i = 0; i < result.length; i++) {
      int firstDigit = _ALPHABET.indexOf(str[i * 2]);
      int secondDigit = _ALPHABET.indexOf(str[i * 2 + 1]);
      if (firstDigit == -1 || secondDigit == -1) {
        throw new FormatException("Non-hex character detected in $hex");
      }
      result[i] = (firstDigit << 4) + secondDigit;
    }
    return result;
  }

  @override
  Widget build(BuildContext context) {
    var actions = Row(
      mainAxisSize: MainAxisSize.min,
      children: <Widget>[
        IconButton(
          icon: Icon(
            Icons.cloud_download,
            color: Theme.of(context).iconTheme.color.withOpacity(0.5),
          ),
          onPressed: _onReadPressed,
        ),
        IconButton(
          icon: Icon(Icons.cloud_upload,
              color: Theme.of(context).iconTheme.color.withOpacity(0.5)),
          onPressed: () async {
            final String txt = await _asyncInputDialog(context);
            _onWritePressed(txt);
          },
        ),
        IconButton(
          icon: Icon(Icons.sync,
              color: Theme.of(context).iconTheme.color.withOpacity(0.5)),
          onPressed: _onNotificationON,
        ),
        IconButton(
          icon: Icon(Icons.sync_disabled,
              color: Theme.of(context).iconTheme.color.withOpacity(0.5)),
          onPressed: _onNotificationOFF,
        ),
      ],
    );
    var title = Column(
      mainAxisAlignment: MainAxisAlignment.center,
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        const Text('Characteristic'),
        Text(characteristic.uuid,
            style: Theme.of(context)
                .textTheme
                .body1
                .copyWith(color: Theme.of(context).textTheme.caption.color))
      ],
    );

    return ListTile(
      title: title,
      trailing: actions,
    );
  }
}
